@@ -10,6 +10,7 @@ def clean_literal_string(literal_term: str) -> str:
     new_literal_term = re.sub(r"\^\^\w+:\w+", "", new_literal_term)
     return new_literal_term
 
+
 # TODO: set supppression key as constant
 def remove_suppression_key(term: str) -> str:
     return term.replace("*", "")
@@ -24,8 +25,13 @@ def get_term_aliases(term: str) -> list[str]:
     return []
 
 
-def format_literal(literal: Literal) -> str:
-    literal_str = f'"{literal.value}"'
+def format_literal(literal: Literal, prefix: str) -> str:
+    if prefix is None:
+        raise ValueError(
+            "The literal datatype prefix was not specified. Literal datatype namespaces cannot be None."
+        )
+    literal_value = literal.value if literal.value else str(literal)
+    literal_str = f'"{literal_value}"'
     lang_str = (
         f"@{literal.language}"
         if hasattr(literal, "language") and literal.language
@@ -36,6 +42,6 @@ def format_literal(literal: Literal) -> str:
     if hasattr(literal, "datatype") and literal.datatype:
         datatype = literal.datatype
         _, abbrev = split_uri(datatype)
-        datatype_str = f"^^xsd:{abbrev}"
+        datatype_str = f"^^{prefix}:{abbrev}"
 
     return f"{literal_str}{lang_str}{datatype_str}"
